#include "Vec.h"
#include "Mat.h"
#include "Image.h"
#include "misc.h"
#include "SmoothingFilter.h"
#include "Gauss.h"
#include "FeaturePoint.h"

extern double DET_EPS;

double
FeaturePoint::selfSSD(const Image &im,int x, int y, const Gauss &gauss) const
{
   double J = 0.0;
   double i_p, i_xy;
   int width=im.xl(), height=im.yl();
   int range = gauss.range;
   VecC p_;
  
   p_[0] = point[0]+x; p_[1] = point[1]+y;
   if(!is_in_image(p_, range, range, width - range, height - range))
      return -1;
  
   for(int i=-range; i<range ; i++){
      for(int j=-range; j<range; j++){
         i_p  = (double)im.rget((int)point[0] + i,     (int)point[1] + j);
         i_xy = (double)im.rget((int)point[0] + i + x, (int)point[1] + j + y);
         J   += gauss.table[i+range][j+range] * (i_p -i_xy)*(i_p - i_xy);
      }
   }		
   return J;
}
	
double
FeaturePoint::partial_differential(const Image &im,
                                   const SmoothingFilter &mask,
                                   const VecC&p)
{
   double tp_index[2];//(u,v濴Ȥ)ʬκɸ(x,yɸ)
   double value;
   double ret = 0.0;
   int xcenter = im.xl()/2;
   int ycenter = im.yl()/2;

   int f_size = mask.get_filter_size();

   for(int i = -f_size; i <= f_size; i++){
      tp_index[1] = (p[0] + i) - xcenter;
      for(int j= -f_size; j <= f_size; j++){
         tp_index[0] = ycenter - (p[1] + j);
         value = im.realget(tp_index[0], tp_index[1]);
         ret += mask.get(i,j) * value;
      }
   }
   return ret;
}


FeaturePoint::FeaturePoint(): point(3), cov(2,2), Axis(2,2), T(2,2)
{
   point[0] = point[1] = point[2] = 0;
   cov.clear();
   Axis[0][1] = Axis[1][0] = 0;
   Axis[0][0] = Axis[1][1] = 1;
   T =  Axis;
   be=false;
}

//FeaturePoint::~FeaturePoint(){} //إå
		
bool
FeaturePoint::calc_parameter()
{
   VecC eigen(2);
   eigen_vector(cov, T, eigen);
   if(eigen[0]<0 || eigen[1] < 0){
      cerr << point << " is eliminate.(eigen value is negative)"<<endl;
      return false;
   }
  
   //ľι󼰤ΤȤϡͭ٥ȥ촹
   if(det(T) < 0){
      double tmp;
      tmp = T[0][0]; T[0][0]= T[0][1]; T[0][1] = tmp;
      tmp = T[1][0]; T[1][0]= T[1][1]; T[1][1] = tmp;			
      tmp = eigen[0]; eigen[0] = eigen[1]; eigen[1] = tmp;
   }
   Axis[0][0] = sqrt(eigen[0]);Axis[1][1] = sqrt(eigen[1]);
  
   //180٤礭žϵղžˤ
   if(T[0][0] < 0) T = -T;
   return true;
}
	
//ʻĺˤ붦ʬ׻롣
bool
FeaturePoint::calc_cov_by_residual(const Image &im, const Gauss& gauss)
{
   Mat A;
   VecC b, n;
   A.clear();b.clear();
   int width = im.xl(), height = im.yl();
   int range = gauss.range;
   if(!is_in_image(point,range, range, width - range, height - range))
      return false;
   for(int i = -range; i<range; i++){
      for(int j=-range; j<range; j++){
         VecC m( (double)i*i, 2.0*i*j, (double)j*j );
         double ssd = selfSSD(im,i,j,gauss);
         if(ssd < 0) return false;
  
         A += gauss.table[i+range][j+range] * m * (~m);
         b += gauss.table[i+range][j+range] * ssd * m;					
      }
   }
		
   n = inverse( (~A)*A ) * (~A) * b;
   double deter = n[0]*n[2] - n[1]*n[1];
   if(fabs(deter) < DET_EPS){
      cerr <<"covariance can't be calculated. becouse determinant is 0."<<endl;
      return false;
   }
   Mat V(2,2);
   V[0][0] = n[0]; V[0][1] = n[1];
   V[1][0] = n[1]; V[1][1] = n[2];
   cov = inverse(V);
/*  cov[0][0] =  n[2]/deter; cov[0][1] = -n[1]/deter;
    cov[1][0] = -n[1]/deter; cov[1][1] = n[0]/deter;*/
   return true;
}

//ʬˤ붦ʬη׻
bool
FeaturePoint::calc_cov_by_derivative(const Image &im,
                                     const SmoothingFilter &mask_u,
                                     const SmoothingFilter &mask_v,
                                     int template_size)
{
   VecC NbI(2);       // ƥ顼Ÿ򤷤ʬͤ
   Mat A(2,2);        // إå
   VecC index(2);  // μɸ
   int filter_size = mask_u.get_filter_size();
   int width = im.xl(), height = im.yl();
   if( ! is_in_image( point, filter_size + template_size,       //
                      filter_size + template_size,              //
                      width - (filter_size + template_size),    //
                      height - (filter_size + template_size) ) )//
      return false;  // out of image


   A.clear(0.0);
   for(int j = -template_size; j <= template_size; j++){
      index[1] = point[1] + j;
      for(int i = -template_size; i <= template_size; i++){
         index[0] = point[0] + i;
         NbI[0] = partial_differential(im, mask_u, index);//ʬ
         NbI[1] = partial_differential(im, mask_v, index);//ļʬ
         A += NbI * (~NbI);
      }
   }
  
   if(fabs( det(A) ) <= EPS){
      cerr <<"covariance matrix is singular"<<endl;
      return false;
   }
   cov = inverse(A);
   return true;
}
