#include <iostream>

using std::cerr;
using std::endl;
using std::cout;

#include <math.h>

#include "Vec.h"
#include "Mat.h"
#include "TriVecDesc.h"
#include "FeaturePoint.h"
#include "Gauss.h"
#include "misc.h"    //is_in_image()

extern double DET_EPS;


//ץե٥ȥʬ
bool
TriVector::make_profile_vector(const Image & im, 
                               const FeaturePoint &start,
                               const FeaturePoint& end,
                               //int dim,
                               const Gauss & gauss)
{
   VecC direction = end.point - start.point;
   VecC pt;
  
   for(int i=0; i<dim; i++){
      pt =  direction / (dim -1 ) * i + start.point ;
      profile[i] = gauss.smoothing(im, pt);		
      if(profile[i] < 0)
         return false;
   }
  
   return true;
}
	
	
void
TriVector::calc_Rotation(VecC & _direction, Mat &Rotation)
{
   VecC direction(2); direction[0] = _direction[0]; direction[1] = _direction[1];  
   VecC u(2);
   u[0] = 1; u[1] = 0;
   double cos_theta = (direction,u) / norm(direction);
   double theta     = (direction[1] < 0 )? acos(cos_theta) : -acos(cos_theta);
  
   Rotation[0][0] = cos_theta;   Rotation[0][1] = sin(theta);
   Rotation[1][0] = -sin(theta); Rotation[1][1] = cos_theta; 
}
	
//ɸΰ
bool
TriVector::make_standard_region(double *STReg, const Image & im,
                                const FeaturePoint &pt,const Mat & rotation,
                                int radius_scale,
                                //int radius_div, int angle_div,
                                const Gauss &gauss)
{
   VecC pe(2),pc(2);//ץ󥰺ɸɸ
   VecC t(2); t[0] = pt.point[0]; t[1] = pt.point[1];
   int range=gauss.range;
   int width  = im.xl();
   int height = im.yl();
   for (int r=1; r < radius_div+1; r++){
      for(int angle = 0; angle < angle_div; angle++){
         double th = 2 * M_PI * angle / angle_div ;
         pc[0] = (double) r * radius_scale * cos( th );
         pc[1] = (double) r * radius_scale * sin( th );
         pc = rotation * pc;//ץե٥ȥ˹碌ž
  
         pe = ~pt.T * pt.Axis * pt.T * pc + t; //ץ󥰺ɸλ
         if( ! is_in_image(pe, range,range,
                           width  - range, height - range)){
            return false;
         }
         STReg[ (r-1) * angle_div + angle ] = gauss.smoothing(im,pe);
         if(STReg[ (r-1) * angle_div + angle] < 0)
            return false;		
      }
   }
  
   return true;
}
	
//ȥ٥ȥƤǤʿѤ(SSD˹碌)
bool
TriVector::normalize_SSD()
{
   double total = 0.0;
   double avg;
   //profile vector
   for(int i=0; i<dim; i++){
      total += profile[i];
   }
   avg = total / dim;
   for(int i=0; i<dim; i++){
      profile[i] -= avg;
   }
  
   //standard region at start
   total = 0.0;
   for(int i=0; i<radius_div*angle_div; i++){
      total += STRstart[i];
   }
   avg = total / (radius_div*angle_div);
   for(int i=0; i<radius_div*angle_div; i++){
      STRstart[i] -= avg;
   }
  
   //standard region at end
   total=0.0;
   for(int i=0; i<radius_div*angle_div; i++){
      total += STRend[i];
   }
   avg = total / (radius_div*angle_div);
   for(int i=0; i<radius_div*angle_div; i++){
      STRend[i]   -= avg;
   }
   return be; 
}

//ȥ٥ȥƤǤ2Υ(ؤ˹碌)
bool
TriVector::normalize_CC()
{
   double nrm = 0.0;
   for(int i=0; i<dim; i++){
      nrm += profile[i]*profile[i];	
   }
   for(int i=0; i<radius_div*angle_div; i++){
      nrm += STRstart[i]*STRstart[i];
      nrm += STRend[i]*STRend[i];
   }
   nrm = sqrt(nrm);
   if(fabs(nrm) < DET_EPS){
      be = false;
      return be;
   }
   for(int i=0; i<dim; i++){
      profile[i]  /= nrm;
   }
   for(int i=0; i<radius_div*angle_div; i++){
      STRstart[i] /= nrm;
      STRend[i]   /= nrm;
   }  	
   return be;
}


TriVector::TriVector()
{
   profile = STRstart = STRend = 0;
   dim        = 0;
   radius_div = 0;
   angle_div  = 0;
   be=false;
}

TriVector::TriVector(const TriVector &tv)
{
   dim        = tv.dim;
   radius_div = tv.radius_div;
   angle_div  = tv.angle_div;
   be         = tv.be;
   if(!be)return ;
   profile  = new double[dim];
   STRstart = new double[angle_div*radius_div];
   STRend   = new double[angle_div*radius_div];
   if(!profile || !STRstart ||! STRend){
      cerr <<"can't get memory TriVector copy_cnstr"<<endl;
      exit(1);		
   }
   for(int i=0; i<dim; i++){
      profile[i] = tv.profile[i];
   }
   for(int i=0; i<radius_div*angle_div; i++){
      STRstart[i] = tv.STRstart[i];
      STRend[i]   = tv.STRend[i];
   }
  
}

TriVector& TriVector::operator=(const TriVector &tv)
{
   if (this == &tv) return *this;
   assert(dim        == tv.dim);
   assert(angle_div  == tv.angle_div);
   assert(radius_div == tv.radius_div);
	  
   be         = tv.be;
   if(!be) return *this ;
   if(!profile)
      profile  = new double[dim];
   if(!STRstart)
      STRstart = new double[angle_div*radius_div];
   if(!STRend)
      STRend   = new double[angle_div*radius_div];
   if(!profile || !STRstart ||! STRend){
      cerr <<"can't get memory TriVector copy_cnstr"<<endl;
      exit(1);		
   }
  
   for(int i=0; i<dim; i++){
      profile[i] = tv.profile[i];
   }
   for(int i=0; i<radius_div*angle_div; i++){
      STRstart[i] = tv.STRstart[i];
      STRend[i]   = tv.STRend[i];
   }
   return *this;
}


TriVector::~TriVector()
{
   if(profile)
      delete [] profile;
   if(STRstart)
      delete [] STRstart;
   if(STRend)
      delete [] STRend;
   profile = STRstart = STRend = 0;
}
  
//2֤λȥ٥ȥ
bool
TriVector::make_tri_vector(const Image & im,const Gauss &gauss,
                           const FeaturePoint &start,const FeaturePoint &end,
                           int dimension,double d_min, double d_max,
                           int radius_division, int radius_scale, 
                           int angle_division)
{
   Mat rotation(2,2);
   VecC direction = end.point - start.point;
   double distance = (direction, direction);//2躬Ѥ
   if(distance < d_min * d_min || d_max * d_max < distance){
      be = false;
      return be;
   }
   be         = true;
   dim        = dimension;
   radius_div = radius_division;
   angle_div  = angle_division;
   profile  = new double[dim];
   STRstart = new double[radius_div * angle_div];
   STRend   = new double[radius_div * angle_div];
   if(!profile || !STRstart || !STRend){
      cerr << "can't get memory.make_tri-vector()."<<endl;
      exit(1);
   }
   if ( !make_profile_vector(im, start, end, gauss) ){
      be = false;
   }
   calc_Rotation(direction, rotation);	
   if ( !make_standard_region(STRstart, im, start, rotation, radius_scale, gauss) ){
      be = false;
   }
   if ( !make_standard_region(STRend,   im,  end,  rotation, radius_scale, gauss) ){
      be = false;
   }
  
   if(!be){
      dim = radius_div = angle_div = 0;
      delete [] profile;
      delete [] STRstart;
      delete [] STRend;
      profile = STRstart = STRend = NULL;	
   }
   return be;
}

//SSDˤٷ׻
//tv1,tv2  : ٤η׻оݤȤʤ뻰ȥ٥ȥ
//pv_th    : ξץե٥ȥγǤкФ []
//str_th   : ξɸΰ٥ȥγǤкФ []
double
TriVector::similarity_SSD(const TriVector &tv1, const TriVector &tv2,
                          double pv_th, double str_th,
                          double pv_weight, double str_weight)
{
   double pv_ssd=0.0;
   double strs_ssd=0.0;
   double stre_ssd=0.0;
   double similarity=0.0;
  
   double dimension=tv1.dim;
   static double str_dim = tv1.radius_div * tv1.angle_div;
   static double pv_th2  = pv_th * pv_th;
   static double str_th2 = str_th * str_th;
   //ץե٥ȥʬ
   for(int i=0; i<dimension; i++){
      pv_ssd += (tv1.profile[i] - tv2.profile[i]) * (tv1.profile[i] - tv2.profile[i]);
   }
   pv_ssd /= dimension;
   if(pv_ssd > pv_th2)
      return 0.0;
  
   //ɸΰ٥ȥʬ
   for(int i=0; i<str_dim; i++){
      strs_ssd += (tv1.STRstart[i] - tv2.STRstart[i]) * (tv1.STRstart[i] - tv2.STRstart[i]);
   }
   strs_ssd /= str_dim;
   if(strs_ssd > str_th2)
      return 0.0;

   //ɸΰ٥ȥʬ
   for(int i=0; i<str_dim; i++){
      stre_ssd += (tv1.STRend[i] - tv2.STRend[i]) * (tv1.STRend[i] - tv2.STRend[i]);
   }
   stre_ssd /= str_dim;
   if(stre_ssd > str_th2)
      return 0.0;
  
   double sim_pv  = 1 - pv_ssd / pv_th2;
   double sim_str = 1 - (strs_ssd + stre_ssd) / (2*str_th2);
   //SSD٤ؤѴ
//   similarity = 3 - ( pv_ssd / pv_th2 + (strs_ssd + stre_ssd) / str_th2);
   similarity = pv_weight*sim_pv + str_weight*sim_str; //4 ϽŤ
   return similarity  ;
}
