#include <iostream>

using std::cerr;
using std::endl;

#include "Mat.h"
#include "Vec.h"
#include "Image.h"
#include "FeaturePoint.h"
#include "Gauss.h"
#include "TriangleVector.h"

TriangleVector::TriangleVector(const Image &img, const Gauss &gauss,
                               int length, 
                               const FeaturePoint &pt1,
                               const FeaturePoint &pt2,
                               const FeaturePoint &pt3)
{
   triangle=0;dim=0;
   Mat A(2,2);
   VecC edge1(pt2.point - pt1.point), edge2(pt3.point - pt1.point);
   A[0][0] = edge1[0] / length; A[0][1] = edge2[0] / length;
   A[1][0] = edge1[1] / length; A[1][1] = edge2[1] / length;
   if(fabs(A[0][0]*A[1][1] - A[0][1]*A[1][0]) < DET_EPS){ //31ľ˾äƤ
      be = false;
      return ;
   }else be = true;

   dim = (length -2)*(length-3);
   triangle = new double [dim];
   if(!triangle){
      cerr << "can't get memory TriangleVector()"<<endl;
      exit(1);
   }
   int count = 0;
   VecC t(2);t[0] = pt1.point[0]; t[1] = pt1.point[1];
   for(int i=1; i<length -1; i++){
      for(int j=1; j < length -i; j++){
         VecC sampling_point(2),tmp(2);tmp[0] = i; tmp[1] = j;
         sampling_point = A*tmp + t;
         triangle[count] = gauss.smoothing(img, sampling_point);
         if(triangle[count] < 0){ //ϤߤǤ
            be = false;
            return;
         }
         count++;
      }
   }
   return ;
}
	
TriangleVector::~TriangleVector()
{
   if(triangle)delete[] triangle;
}

double
TriangleVector::similarity_SSD(const TriangleVector &tr1,
                               const TriangleVector &tr2,
                               double triangle_th,
                               double triangle_weight)
{
   int dimension = tr1.dim;
   static double th2 = triangle_th * triangle_th;
   double ssd = 0.0;
  
   for(int i=0; i<dimension; i++){		
      ssd += (tr1.triangle[i] -  tr2.triangle[i])*(tr1.triangle[i] -  tr2.triangle[i]);
   }
   ssd /= dimension;
   if(ssd > th2) return 0.0;
   double similarity = 1 - (ssd / th2);
   return similarity * triangle_weight;
}
