#if __GNUC__ < 3
# include <iostream.h>
#else
# include <iostream>
using std::ostream;
#endif

#include <math.h>
#include <Mat.h>
#include "Quaternion.h"

// ----------------------------------------------------------------------
// 󥹥ȥ饯
// ----------------------------------------------------------------------
Quaternion::Quaternion()
{
    setIdentity();
}
// ----------------------------------------------------------------------
Quaternion::Quaternion(double s, double v0, double v1, double v2)
{
    this->s = s;
    this->v[0] = v0;
    this->v[1] = v1;
    this->v[2] = v2;
}
// ----------------------------------------------------------------------
Quaternion::Quaternion(double s, VecR v)
{
    this->s = s;
    this->v = v;
}
// ----------------------------------------------------------------------
Quaternion::Quaternion(double s, VecC v)
{
    this->s = s;
    this->v = ~v;
}
// ----------------------------------------------------------------------
// 󥹥ȥ饯(󤫤4)
// ----------------------------------------------------------------------
Quaternion::Quaternion(const Mat &m) 
{ 
    double tr, w; 

    tr = m[0][0] + m[1][1] + m[2][2]; 
    
    if (tr >= 0) 
    { 
	w = sqrt(tr + 1); 
	this->s = 0.5 * w; 
	w = 0.5 / w; 
	this->v[0] = (m[2][1] - m[1][2]) * w; 
	this->v[1] = (m[0][2] - m[2][0]) * w; 
	this->v[2] = (m[1][0] - m[0][1]) * w; 
    }
    else
    { 
	int i; 
	double max; 

	i = 0;
	max = m[0][0]; 
	if (m[1][1] > max)
	{
	    i = 1;
	    max = m[1][1];
	} 
	if (m[2][2] > max)
	{
	    i = 2;
	} 

	switch (i)
	{ 
	case 0: 
	    w = sqrt(m[0][0] - m[1][1] - m[2][2] + 1); 
	    this->v[0] = 0.5 * w; 
	    w = 0.5 / w; 
	    this->v[1] = (m[0][1] + m[1][0]) * w; 
	    this->v[2] = (m[2][0] + m[0][2]) * w; 
	    this->s = (m[2][1] - m[1][2]) * w; 
	    break; 
	case 1: 
	    w = sqrt(m[1][1] - m[2][2] - m[0][0] + 1); 
	    this->v[1] = 0.5 * w; 
	    w = 0.5 / w; 
	    this->v[2] = (m[1][2] + m[2][1]) * w; 
	    this->v[0] = (m[0][1] + m[1][0]) * w; 
	    this->s = (m[0][2] - m[2][0]) * w; 
	    break; 
	case 2: 
	    w = sqrt(m[2][2] - m[0][0] - m[1][1] + 1); 
	    this->v[2] = 0.5 * w; 
	    w = 0.5 / w; 
	    this->v[0] = (m[2][0] + m[0][2]) * w; 
	    this->v[1] = (m[1][2] + m[2][1]) * w; 
	    this->s = (m[1][0] - m[0][1]) * w; 
	    break; 
	} 
    } 
}
// ----------------------------------------------------------------------
// ǥȥ饯
// ----------------------------------------------------------------------
Quaternion::~Quaternion()
{
}
// ----------------------------------------------------------------------
// ñ4ˤ
// ----------------------------------------------------------------------
void Quaternion::setIdentity()
{
    this->s = 1;
    this->v[0] = 0;
    this->v[1] = 0;
    this->v[2] = 0;
}
// ----------------------------------------------------------------------
// 4
// ----------------------------------------------------------------------
Quaternion& Quaternion::operator =(const Quaternion& Q)
{
    this->s = Q.s;
    this->v = Q.v;

    return *this;
}
// ----------------------------------------------------------------------
// 4Ѵ
// ----------------------------------------------------------------------
Mat Quaternion::getMat()
{
    Mat m;
    m[0][0] = 1 - 2 * v[1] * v[1] - 2 * v[2] * v[2];
    m[0][1] = 2 * v[0] * v[1] - 2 * s * v[2];
    m[0][2] = 2 * v[0] * v[2] + 2 * s * v[1];
    m[1][0] = 2 * v[0] * v[1] + 2 * s * v[2];
    m[1][1] = 1 - 2 * v[0] * v[0] - 2 * v[2] * v[2];
    m[1][2] = 2 * v[1] * v[2] - 2 * s * v[0];
    m[2][0] = 2 * v[0] * v[2] - 2 * s * v[1];
    m[2][1] = 2 * v[1] * v[2] + 2 * s * v[0];
    m[2][2] = 1 - 2 * v[0] * v[0] - 2 * v[1] * v[1];

    return m;
}
// ----------------------------------------------------------------------
// vž
// ----------------------------------------------------------------------
VecC Quaternion::rotateVec(const VecC &v)
{
    VecC vd = ~((*this) * Quaternion(0, v) * (*this).inverse()).getVecR();
    return vd / vd.norm();
}
// ----------------------------------------------------------------------
// axisǳanglež4򥻥åȤ
// ----------------------------------------------------------------------
void Quaternion::setRotate(const VecR &axis, double angle)
{
    s = cos(-angle / 2);
    v = axis * sin(-angle / 2);
}
// ----------------------------------------------------------------------
// axisǳanglež4򥻥åȤ
// ----------------------------------------------------------------------
void Quaternion::setRotate(const VecC &axis, double angle)
{
    s = cos(-angle / 2);
    v = ~axis * sin(-angle / 2);
}
// ----------------------------------------------------------------------
// տ֤
// ----------------------------------------------------------------------
Quaternion Quaternion::inverse()
{
    return conjugate() / norm();
}
// ----------------------------------------------------------------------
// Τ֤
// ----------------------------------------------------------------------
Quaternion Quaternion::normalize()
{
    return (*this) / norm();
}
// ----------------------------------------------------------------------
// ֤
// ----------------------------------------------------------------------
Quaternion Quaternion::conjugate()
{
    return Quaternion(s, -v);
}
// ----------------------------------------------------------------------
// Ĺ֤
// ----------------------------------------------------------------------
double Quaternion::norm()
{
    return sqrt(s * s + (v, v));
}
// ----------------------------------------------------------------------
// SLERP:
// ----------------------------------------------------------------------
Quaternion slerp(const Quaternion& q1, const Quaternion& q2, double t)
{
    assert(t >= 0 && t <= 1);

    double theta = acos((q1, q2));
    if (theta == 0.0) //Ʊʤ׻ʤ
	return q1;

    //assert(q1.v != -q2.v); // ȿФ

    return ((q1 * sin(theta * (1 - t))) + (q2 * sin(theta * t))) / sin(theta);
}
// ----------------------------------------------------------------------
// SQUAD:̻շ
// ----------------------------------------------------------------------
Quaternion squad(const Quaternion &q1, const Quaternion& b1, 
	   const Quaternion& b2, const Quaternion& q2, double t)
{
    assert(t >= 0 && t <= 1);

    return slerp(slerp(q1, q2, t), slerp(b1, b2, t), 2 * t * (1 - t));
}
// ----------------------------------------------------------------------
// Ѥ׻
// ----------------------------------------------------------------------
double operator ,(const Quaternion& q1, const Quaternion& q2)
{
   return (q1.v, q2.v);
}
// ----------------------------------------------------------------------
// Q + Q û
// ----------------------------------------------------------------------
Quaternion operator +(const Quaternion &q1, const Quaternion &q2)
{
    return Quaternion(q1.s + q2.s, q1.v + q2.v);
}
// ----------------------------------------------------------------------
// Q * Q 軻
// ----------------------------------------------------------------------
Quaternion operator *(const Quaternion &q1, const Quaternion &q2)
{
    return Quaternion(q1.s * q2.s - (q1.v, q2.v), 
		q1.s * q2.v + q2.s * q1.v + (q1.v % q2.v));
}
// ----------------------------------------------------------------------
// Q * s 軻
// ----------------------------------------------------------------------
Quaternion operator *(const Quaternion &Q, double d)
{
    return Quaternion(Q.s * d, Q.v * d);
}
// ----------------------------------------------------------------------
// Q / s 
// ----------------------------------------------------------------------
Quaternion operator /(const Quaternion &Q, double d)
{
    return Quaternion(Q.s / d, Q.v / d);
}
// ----------------------------------------------------------------------
// ϥڥ졼
// ----------------------------------------------------------------------
ostream& operator << (ostream& s, const Quaternion& Q)
{
   s << "[ " << Q.s << ", " << Q.v << " ]";
   
   return s;
}
// ----------------------------------------------------------------------

#if 0
// ----------------------------------------------------------------------
// exp֤(Υब1ΤȤʳʤ)
// ----------------------------------------------------------------------
Quaternion Quaternion::exp()         
{
    double vn = v.norm();
    return Quaternion(cos(vn), sin(vn) * v / (vn));
}
// ----------------------------------------------------------------------
// log֤(Υब1ΤȤʳʤ)
// ----------------------------------------------------------------------
Quaternion Quaternion::log()
{
    return Quaternion(0, atan(v.norm() / s) * v);
}
// ----------------------------------------------------------------------
// ֤ͤ
// ----------------------------------------------------------------------
double Quaternion::abs()
{
    return sqrt(Norm());
}
#endif
